unit FrmMain;

interface

uses

  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants, System.UIConsts, System.IniFiles,

  FMX.Types, FMX.Controls, FMX.Forms, FMX.Graphics, FMX.Dialogs, System.Math.Vectors, FMX.Controls3D, FMX.Objects3D,
  FMX.Viewport3D, FMX.Controls.Presentation, FMX.StdCtrls, FMX.Layers3D, FMX.Layouts, FMX.ExtCtrls, FMX.Objects,
  FMX.Edit, FMX.EditBox, FMX.SpinBox,

  CubeMath;

type
  TForm3DMain = class(TForm)
    Panel1: TPanel;
    Viewport3D: TViewport3D;
    CubeView: TDummy;
    CubeFace0: TImage3D;
    CubeFace5: TImage3D;
    CubeFace4: TImage3D;
    CubeFace2: TImage3D;
    CubeFace3: TImage3D;
    CubeFace1: TImage3D;
    TimerAnimation: TTimer;
    LabelGenerations: TLabel;
    ButtonExit: TButton;
    CheckBoxShowNumbers: TCheckBox;
    OpenDialogSettings: TOpenDialog;
    ButtonLoad: TButton;
    ButtonReset: TButton;
    ButtonRunStop: TButton;
    TrackBarX: TTrackBar;
    TrackBarZ: TTrackBar;
    TrackBarY: TTrackBar;
    Label2: TLabel;
    TrackBarSpeed: TTrackBar;
    SpinBoxDecimation: TSpinBox;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Label8: TLabel;
    EditRuleStrGreen: TEdit;
    EditRuleStrRed: TEdit;
    ButtonSave: TButton;
    SaveDialogSettings: TSaveDialog;
    CheckBoxUseDecimation: TCheckBox;
    CheckBoxUseRedRule: TCheckBox;
    CheckBoxUse1: TCheckBox;
    CheckBoxUseBackground: TCheckBox;
    ButtonClear: TButton;
    procedure TimerAnimationTimer(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ButtonExitClick(Sender: TObject);
    procedure CheckBoxShowNumbersChange(Sender: TObject);
    procedure FormPaint(Sender: TObject; Canvas: TCanvas; const ARect: TRectF);
    procedure CubeFace0LayerMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Single);
    procedure ButtonLoadClick(Sender: TObject);
    procedure ButtonResetClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure ButtonRunStopClick(Sender: TObject);
    procedure TrackBarXChange(Sender: TObject);
    procedure TrackBarYChange(Sender: TObject);
    procedure TrackBarZChange(Sender: TObject);
    procedure TrackBarSpeedChange(Sender: TObject);
    procedure SpinBoxDecimationChange(Sender: TObject);
    procedure CheckBoxUseBackgroundChange(Sender: TObject);
    procedure CheckBoxUseRedRuleChange(Sender: TObject);
    procedure CheckBoxUse1Change(Sender: TObject);
    procedure CheckBoxUseDecimationChange(Sender: TObject);
    procedure ButtonSaveClick(Sender: TObject);
    procedure ButtonClearClick(Sender: TObject);
    procedure EditRuleStrGreenChange(Sender: TObject);
    procedure EditRuleStrRedChange(Sender: TObject);
  private
    FCube           : TCubeLife;
    FDecimation     : Integer;
    FAniCounter     : Integer;
    FSpeed          : Single;
    FSavedSpeed     : Single;
    FGenerations    : UInt64;
    FShowNumbers    : Boolean;
    FUseBackground  : Boolean;
    FUseRedRule     : Boolean;
    FUse1ForMissing : Boolean;
    FUseDecimation  : Boolean;
    FRuleStrGreen   : string;
    FRuleStrRed     : string;
    FLatestFile     : string;
    FRunning        : Boolean;
    FRotX           : Single;
    FRotY           : Single;
    FRotZ           : Single;
  private
    procedure   SetGenerations   ( aValue: UInt64);
    procedure   SetShowNumbers   ( aValue: Boolean);
    procedure   SetUseBackground ( aValue: Boolean);
    procedure   SetUseRedRule    ( aValue: Boolean);
    procedure   SetUse1ForMissing( aValue: Boolean);
    procedure   SetUseDecimation ( aValue: Boolean);
    procedure   SetRuleStrGreen  ( const aValue: string);
    procedure   SetRuleStrRed    ( const aValue: string);
    procedure   SetLatestFile( const aValue: string);
    procedure   SetDecimation( aValue: Integer);
    procedure   SetSpeed( aValue: Single);
    procedure   SetRunning( aValue: Boolean);
    procedure   SetRotX( aValue: Single);
    procedure   SetRotY( aValue: Single);
    procedure   SetRotZ( aValue: Single);
  private
    procedure   RotateCubeViewX( anAmount: Single);
    procedure   RotateCubeViewY( anAmount: Single);
    procedure   RotateCubeViewZ( anAmount: Single);
    procedure   RotateCubeView ( anAmount: Single);
    procedure   UpdateView;
    procedure   AnimateFaces;
    procedure   TimerTick;
    procedure   InitCube;
    procedure   DoneCube;
    procedure   ResetCube;
    procedure   ClearCube;
    procedure   LoadFile( const aFileName: string);
    procedure   SaveFile( const aFileName: string);
    function    ExeName    : string;
    function    IniFileName: string;
    function    BackName   : string;
    procedure   LoadIni;
    procedure   SaveIni;
    procedure   LoadSettings;
    procedure   SaveSettings;
    procedure   HandleMouseClick( const aSender: TImage3D; aButton: TMouseButton; aShift: TShiftState; X, Y: Single);
    function    GetBackgroundName( const aSender: TObject): string;
  private
    property    Generations    : UInt64  read FGenerations    write SetGenerations;
    property    ShowNumbers    : Boolean read FShowNumbers    write SetShowNumbers;
    property    UseRedRule     : Boolean read FUseRedRule     write SetUseRedRule;
    property    Use1ForMissing : Boolean read FUse1ForMissing write SetUse1ForMissing;
    property    UseDecimation  : Boolean read FUseDecimation  write SetUseDecimation;
    property    RuleStrGreen   : string  read FRuleStrGreen   write SetRuleStrGreen;
    property    RuleStrRed     : string  read FRuleStrRed     write SetRuleStrRed;
    property    UseBackground  : Boolean read FUseBackground  write SetUseBackground;
    property    LatestFile     : string  read FLatestFile     write SetLatestFile;
    property    Decimation     : Integer read FDecimation     write SetDecimation;
    property    Speed          : Single  read FSpeed          write SetSpeed;
    property    Running        : Boolean read FRunning        write SetRunning;
    property    RotX           : Single  read FRotX           write SetRotX;
    property    RotY           : Single  read FRotY           write SetRotY;
    property    RotZ           : Single  read FRotZ           write SetRotZ;
  end;

var

  Form3DMain: TForm3DMain;


implementation


const

  StrSettings = 'settings';

// User area

  function  FloatMod( const X, Y: Single): Single; inline;
  begin
    Result := Frac( X / Y) * Y;
  end;


  function  MathFloatMod( const X, Y: Single): Single; inline;
  begin
    Result := FloatMod( FloatMod( X, Y) + Y, Y);
  end;


//  private

    procedure   TForm3DMain.SetGenerations( aValue: UInt64);
    begin
      if   aValue >= aValue.MaxValue
      then aValue := 0;

      if   aValue <> FGenerations
      then begin
        FGenerations          := aValue;
        LabelGenerations.Text := UIntToStr( aValue);
      end;
    end;


    procedure   TForm3DMain.SetShowNumbers( aValue: Boolean);
    begin
      if   aValue <> FShowNumbers
      then begin
        FShowNumbers                  := aValue;
        CheckBoxShowNumbers.IsChecked := aValue;
        FCube.ShowNumbers             := aValue;
        UpdateView;
      end;
    end;


    procedure   TForm3DMain.SetUseBackground( aValue: Boolean);
    begin
      if   aValue <> FUseBackground
      then begin
        FUseBackground                  := aValue;
        CheckBoxUseBackground.IsChecked := aValue;
        UpdateView;
      end;
    end;


    procedure   TForm3DMain.SetUseRedRule( aValue: Boolean);
    begin
      if   aValue <> FUseRedRule
      then begin
        FUseRedRule                  := aValue;
        CheckBoxUseRedRule.IsChecked := aValue;
        FCube.UseRedRule             := aValue;
        UpdateView;
      end;
    end;


    procedure   TForm3DMain.SetUse1ForMissing( aValue: Boolean);
    begin
      if   aValue <> FUse1ForMissing
      then begin
        FUse1ForMissing        := aValue;
        CheckBoxUse1.IsChecked := aValue;
        FCube.Use1ForMissing   := aValue;
        UpdateView;
      end;
    end;


    procedure   TForm3DMain.SetUseDecimation( aValue: Boolean);
    begin
      if   aValue <> FUseDecimation
      then begin
        FUseDecimation                  := aValue;
        CheckBoxUseDecimation.IsChecked := aValue;
        UpdateView;
      end;
    end;


    procedure   TForm3DMain.SetRuleStrGreen( const aValue: string);
    begin
      if   aValue <> FRuleStrGreen
      then begin
        FRuleStrGreen         := aValue;
        EditRuleStrGreen.Text := aValue;
        FCube.RuleStrGreen    := aValue;
      end;
    end;


    procedure   TForm3DMain.SetRuleStrRed( const aValue: string);
    begin
      if   aValue <> FRuleStrRed
      then begin
        FRuleStrRed         := aValue;
        EditRuleStrRed.Text := aValue;
        FCube.RuleStrRed    := aValue;
      end;
    end;


    procedure   TForm3DMain.SetLatestFile( const aValue: string);
    begin
      if   aValue <> FLatestFile
      then begin
        FLatestFile := aValue;
        Caption     := 'Cuby 3D: ' + aValue;
      end;
    end;


    procedure   TForm3DMain.SetDecimation( aValue: Integer);
    begin
      if   aValue <> FDecimation
      then Begin
        FDecimation             := aValue;
        SpinBoxDecimation.Value := aValue;
      end;
    end;


    procedure   TForm3DMain.SetSpeed( aValue: Single);
    begin
      if   aValue <> FSpeed
      then begin
        FSpeed              := aValue;
        TrackBarSpeed.Value := aValue;
      end;
    end;


    procedure   TForm3DMain.SetRunning( aValue: Boolean);
    begin
      if   aValue <> FRunning
      then begin
        FRunning := aValue;

        if   FRunning
        then begin
          Speed := FSavedSpeed;
          ButtonRunStop.Text := 'Stop (&R)'
        end
        else begin
          FSavedSpeed := Speed;
          ButtonRunStop.Text := '&Rotate';
        end;
      end;
    end;


    procedure   TForm3DMain.SetRotX( aValue: Single);
    begin
      if   aValue <> FRotX
      then begin
        FRotX                    := aValue;
        TrackBarX.Value          := aValue;
        CubeView.RotationAngle.X := aValue;
      end;
    end;


    procedure   TForm3DMain.SetRotY( aValue: Single);
    begin
      if   aValue <> FRotY
      then begin
        FRotY                    := aValue;
        TrackBarY.Value          := aValue;
        CubeView.RotationAngle.Y := aValue;
      end;
    end;


    procedure   TForm3DMain.SetRotZ( aValue: Single);
    begin
      if   aValue <> FRotZ
      then begin
        FRotZ                    := aValue;
        TrackBarZ.Value          := aValue;
        CubeView.RotationAngle.Z := aValue;
      end;
    end;


//  private

    procedure   TForm3DMain.RotateCubeViewX( anAmount: Single);
    begin
      CubeView.RotationAngle.X := MathFloatMod( CubeView.RotationAngle.X + anAmount, 360);
      RotX                     := CubeView.RotationAngle.X;
    end;


    procedure   TForm3DMain.RotateCubeViewY( anAmount: Single);
    begin
      CubeView.RotationAngle.Y := MathFloatMod( CubeView.RotationAngle.Y + anAmount, 360);;
      RotY                     := CubeView.RotationAngle.Y;
    end;


    procedure   TForm3DMain.RotateCubeViewZ( anAmount: Single);
    begin
      CubeView.RotationAngle.Z := MathFloatMod( CubeView.RotationAngle.Z + anAmount, 360);;
      RotZ                     := CubeView.RotationAngle.Z;
    end;


    procedure   TForm3DMain.RotateCubeView( anAmount: Single);
    begin
      RotateCubeViewX( anAmount *  1.000);
      RotateCubeViewY( anAmount * -1.712);
      RotateCubeViewZ( anAmount *  0.431);
    end;


    procedure   TForm3DMain.UpdateView;
    begin
      FCube.PaintFaceOn( 0, CubeFace0.Bitmap.Canvas, Rect( 0, 0, CubeFace0.Bitmap.Width, CubeFace0.Bitmap.Height));
      FCube.PaintFaceOn( 1, CubeFace1.Bitmap.Canvas, Rect( 0, 0, CubeFace1.Bitmap.Width, CubeFace1.Bitmap.Height));
      FCube.PaintFaceOn( 2, CubeFace2.Bitmap.Canvas, Rect( 0, 0, CubeFace2.Bitmap.Width, CubeFace2.Bitmap.Height));
      FCube.PaintFaceOn( 3, CubeFace3.Bitmap.Canvas, Rect( 0, 0, CubeFace3.Bitmap.Width, CubeFace3.Bitmap.Height));
      FCube.PaintFaceOn( 4, CubeFace4.Bitmap.Canvas, Rect( 0, 0, CubeFace4.Bitmap.Width, CubeFace4.Bitmap.Height));
      FCube.PaintFaceOn( 5, CubeFace5.Bitmap.Canvas, Rect( 0, 0, CubeFace5.Bitmap.Width, CubeFace5.Bitmap.Height));
    end;


    procedure   TForm3DMain.AnimateFaces;
    begin
      FCube.Process;
      UpdateView;
    end;


    procedure   TForm3DMain.TimerTick;
    begin
      if   FRunning
      then begin
        RotateCubeView( Speed);

        if   UseDecimation
        then begin
          Dec( FAniCounter);

          if  FAniCounter <= 0
          then begin
            AnimateFaces;
            FAniCounter := FDecimation;
            Generations := Generations + 1;
          end;
        end;
      end;
    end;


    procedure   TForm3DMain.InitCube;
    begin
      Caption                   := 'Cuby 3D: default pattern';
      FRunning                  := True;
      Running                   := False;
      FLatestFile               := '-';
      LatestFile                := '';
      Generations               :=  1;
      Generations               :=  0;
      Speed                     :=  0.1;
      FSavedSpeed               :=  0.1;
      FDecimation               :=  2;
      Decimation                :=  1;
      FCube                     := TCubeLife.Create( 2); // 2 is a dummy history count here
      FCube.OnGetbackgroundName := GetBackgroundName;
      UseRedRule                := True;
      FUse1ForMissing           := True;
      Use1ForMissing            := False;
      FUseDecimation            := False;
      UseDecimation             := True;
      FCube.Pattern             := '0,2,0;0,2,2;0,2,3;1,0,0;1,0,1;1,1,0;2,0,0;2,0,1;2,0,3;2,1,0;2,1,2;2,1,3;2,3,0;2,3,1;2,3,3;3,2,0;3,2,2;5,1,1';
      RuleStrGreen              := '0,3,1; 1,0,0; 1,1,0; 1,4,0; 1,5,1; 1,6,0; 1,7,0; 1,8,0';
      RuleStrRed                := '0,2,1; 1,0,0; 1,1,0; 1,3,0';
    end;


    procedure   TForm3DMain.DoneCube;
    begin
      FreeAndNil( FCube);
    end;


    procedure   TForm3DMain.ResetCube;
    begin
      LoadFile( LatestFile);
    end;


    procedure   TForm3DMain.ClearCube;
    begin
      FCube.Clear;
      Generations := 0;
      UpdateView;
    end;


    procedure   TForm3DMain.LoadFile( const aFileName: string);
    begin
      if   FileExists( aFileName)
      then begin
        LatestFile := aFileName;

        with TMemIniFile.Create( aFileName)
        do begin
          try
            FCube.Clear;
            Generations  := 0;
            RuleStrGreen   := ReadString ( StrSettings, 'RuleStrGreen'  , RuleStrGreen  );
            RuleStrRed     := ReadString ( StrSettings, 'RuleStrRed'    , RuleStrRed    );
            FCube.Pattern  := ReadString ( StrSettings, 'LatestPattern' , FCube.Pattern );
            Use1ForMissing := ReadBool   ( StrSettings, 'Use1ForMissing', Use1ForMissing);
            UpdateView;
            Generations := 0;
          finally
            DisposeOf;
          end;
        end;
      end;
    end;


    procedure   TForm3DMain.SaveFile( const aFileName: string);
    begin
      LatestFile := aFileName;

      with TMemIniFile.Create( aFileName)
      do begin
        try
          WriteString ( StrSettings, 'RuleStrGreen'  , RuleStrGreen  );
          WriteString ( StrSettings, 'RuleStrRed'    , RuleStrRed    );
          WriteString ( StrSettings, 'LatestPattern' , FCube.Pattern );
          WriteBool   ( StrSettings, 'Use1ForMissing', Use1ForMissing);
        finally
          UpdateFile;
          DisposeOf;
        end;
      end;
    end;


    function    TForm3DMain.ExeName: string;
    begin
      Result := ParamStr( 0);
    end;


    function    TForm3DMain.IniFileName: string;
    begin
      Result := ChangeFileExt( ExeName, '.ini');
    end;


    function    TForm3DMain.BackName: string;
    begin
      Result := IncludeTrailingPathDelimiter( ExtractFilePath( ExeName)) + 'back.bmp';

      if   ( not FileExists( Result))
      or   ( not UseBackground      )
      then Result := '';
    end;


    procedure   TForm3DMain.LoadIni;
    var
      aFileName : string;
      L         : Integer;
      T         : Integer;
      W         : Integer;
      H         : Integer;
    begin
      with TMemIniFile.Create( IniFileName)
      do begin
        try
          Running       := ReadBool   ( StrSettings, 'Running'      , Running      );
          aFileName     := ReadString ( StrSettings, 'LatestFile'   , LatestFile   );
          Speed         := ReadFloat  ( StrSettings, 'Speed'        , Speed        );
          Decimation    := ReadInteger( StrSettings, 'Decimation'   , Decimation   );
          UseDecimation := ReadBool   ( StrSettings, 'UseDecimation', UseDecimation);
          UseBackground := ReadBool   ( StrSettings, 'UseBackground', UseBackground);
          UseRedRule    := ReadBool   ( StrSettings, 'UseRedRule'   , UseRedRule   );
          RotX          := ReadFloat  ( StrSettings, 'RotX'         , RotX         );
          RotY          := ReadFloat  ( StrSettings, 'RotY'         , RotY         );
          RotZ          := ReadFloat  ( StrSettings, 'RotZ'         , RotZ         );
          L             := ReadInteger( StrSettings, 'Left'         , Left         );
          T             := ReadInteger( StrSettings, 'Top'          , Top          );
          W             := ReadInteger( StrSettings, 'Width'        , Width        );
          H             := ReadInteger( StrSettings, 'Height'       , Height       );

          SetBounds( L, T, W, H);
          LoadFile( aFileName);
          UpdateView;
        finally
          DisposeOf;
        end;
      end;
    end;


    procedure   TForm3DMain.SaveIni;
    begin
      with TMemIniFile.Create( IniFileName)
      do begin
        try
          WriteBool   ( StrSettings, 'Running'      , Running      );
          WriteString ( StrSettings, 'LatestFile'   , LatestFile   );
          WriteFloat  ( StrSettings, 'Speed'        , Speed        );
          WriteInteger( StrSettings, 'Decimation'   , Decimation   );
          WriteBool   ( StrSettings, 'UseDecimation', UseDecimation);
          WriteBool   ( StrSettings, 'UseBackground', UseBackground);
          WriteBool   ( StrSettings, 'UseRedRule'   , UseRedRule   );
          WriteFloat  ( StrSettings, 'RotX'         , RotX         );
          WriteFloat  ( StrSettings, 'RotY'         , RotY         );
          WriteFloat  ( StrSettings, 'RotZ'         , RotZ         );
          WriteInteger( StrSettings, 'Left'         , Left         );
          WriteInteger( StrSettings, 'Top'          , Top          );
          WriteInteger( StrSettings, 'Width'        , Width        );
          WriteInteger( StrSettings, 'Height'       , Height       );
        finally
          UpdateFile;
          DisposeOf;
        end;
      end;
    end;


    procedure   TForm3DMain.LoadSettings;
    begin
      if   OpenDialogSettings.Execute
      then LoadFile( OpenDialogSettings.FileName);
    end;


    procedure   TForm3DMain.SaveSettings;
    begin
      if   SaveDialogSettings.Execute
      then SaveFile( SaveDialogSettings.FileName);
    end;


    procedure   TForm3DMain.HandleMouseClick( const aSender: TImage3D; aButton: TMouseButton; aShift: TShiftState; X, Y: Single);
    var
      i     : Integer;
      aCell : TCubeCell;
      bCell : TCubeCell;
    begin
      aCell := FCube.MouseToFaceCell( aSender.Tag, Round( X * 512 / 400), Round( Y * 512 / 400), aSender.Bitmap.Bounds);

      if   ssCtrl in aShift
      then begin
        for i := 0 to aCell.NeighbourCount - 1
        do begin
          bCell := aCell.Neighbour[ i];
          bCell.Alive := not bCell.Alive;
        end;
      end
      else aCell.Alive := not aCell.Alive;

      UpdateView;
    end;


    function    TForm3DMain.GetBackgroundName( const aSender: TObject): string;
    begin
      if   UseBackground
      then Result := BackName
      else Result := '';
    end;


// Delphi area

{$R *.fmx}
{$R *.Surface.fmx  }
{$R *.Windows.fmx  }
{$R *.Macintosh.fmx  }

procedure TForm3DMain.EditRuleStrGreenChange(Sender: TObject);
begin
  RuleStrGreen := EditRuleStrGreen.Text;
end;

procedure TForm3DMain.EditRuleStrRedChange(Sender: TObject);
begin
  RuleStrRed := EditRuleStrRed.Text;
end;

procedure TForm3DMain.SpinBoxDecimationChange(Sender: TObject);
begin
  Decimation := Round( SpinBoxDecimation.Value);
end;

procedure TForm3DMain.TrackBarSpeedChange(Sender: TObject);
begin
  Speed := TrackBarSpeed.Value;
end;

procedure TForm3DMain.TrackBarXChange(Sender: TObject);
begin
  RotX := TrackBarX.Value;
end;

procedure TForm3DMain.TrackBarYChange(Sender: TObject);
begin
  RotY := TrackBarY.Value;
end;

procedure TForm3DMain.TrackBarZChange(Sender: TObject);
begin
  RotZ := TrackBarZ.Value;
end;

procedure TForm3DMain.ButtonClearClick(Sender: TObject);
begin
  ClearCube;
end;

procedure TForm3DMain.ButtonExitClick(Sender: TObject);
begin
  Close;
end;

procedure TForm3DMain.ButtonLoadClick(Sender: TObject);
begin
  LoadSettings;
end;

procedure TForm3DMain.ButtonResetClick(Sender: TObject);
begin
  ResetCube;
end;

procedure TForm3DMain.ButtonRunStopClick(Sender: TObject);
begin
  Running := not Running;
end;

procedure TForm3DMain.ButtonSaveClick(Sender: TObject);
begin
  SaveSettings;
end;

procedure TForm3DMain.CheckBoxShowNumbersChange(Sender: TObject);
begin
  ShowNumbers := CheckBoxShowNumbers.IsChecked;
end;

procedure TForm3DMain.CheckBoxUse1Change(Sender: TObject);
begin
  Use1ForMissing := CheckBoxUse1.IsChecked;
end;

procedure TForm3DMain.CheckBoxUseBackgroundChange(Sender: TObject);
begin
  UseBackground := CheckBoxUseBackground.IsChecked;
end;

procedure TForm3DMain.CheckBoxUseDecimationChange(Sender: TObject);
begin
  UseDecimation := CheckBoxUseDecimation.IsChecked;
end;

procedure TForm3DMain.CheckBoxUseRedRuleChange(Sender: TObject);
begin
  UseRedRule := CheckBoxUseRedRule.IsChecked;
end;

procedure TForm3DMain.CubeFace0LayerMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Single);
begin
  if   Sender is TImage3D
  then HandleMouseClick( TImage3D( Sender), Button, Shift, X, Y);
end;

procedure TForm3DMain.FormCreate(Sender: TObject);
begin
  InitCube;
end;

procedure TForm3DMain.FormDestroy(Sender: TObject);
begin
  SaveIni;
  DoneCube;
end;

procedure TForm3DMain.FormPaint(Sender: TObject; Canvas: TCanvas; const ARect: TRectF);
begin
  UpdateView;
end;

procedure TForm3DMain.FormShow(Sender: TObject);
begin
  LoadIni;
end;

procedure TForm3DMain.TimerAnimationTimer(Sender: TObject);
begin
  TimerTick;
end;

end.
